program KISSURESAT;

{$mode objfpc}{$H+}

uses
  {$IFDEF UNIX}{$IFDEF UseCThreads}
  cthreads,
  {$ENDIF}{$ENDIF}
  Classes, Interfaces, ExtCtrls, Windows, Forms, SysUtils, IniFiles, lNetBase,
  lNet, lNetComponents;

type

TKISSSocket = class
private
  ClientSocket: TLTCPComponent;
  procedure SocketOpen;
  procedure SocketClose;
  procedure OnRead(aSocket: TLSocket);
  procedure OnConnect(aSocket: TLSocket);
  procedure OnDisconnect(aSocket: TLSocket);
  procedure OnError(const msg: string; aSocket: TLSocket);
public
  constructor Create;
  destructor Destroy; override;
end;

TKISS = class
private
  in_buf: string;
  host: string;
  port: integer;
  status: byte;
  timer: TTimer;
  procedure Connect;
  procedure Disconnect;
  procedure Decode(frame: string);
  procedure TimerEvent(Sender: TObject);
public
  constructor Create;
  destructor Destroy; override;
end;

const
  APP_VER       = 'KISS Console - v0.01';
  KISS_INACTIVE	= 0;
  KISS_ACTIVE   = 1;
  KISS_TRY_CONN	= 2;
  FEND          = #$C0;
  FESC          = #$DB;
  TFEND         = #$DC;
  TFESC         = #$DD;

var
  AppIni: TIniFile;
  KISSSocket: TKISSSocket;
  MainForm: TForm;
  KISS: TKISS;

procedure URESAT_tlm(data: pointer; len: integer); cdecl; external 'uresat.dll' name 'procesar';
procedure URESAT_ver; cdecl; external 'uresat.dll' name 'version';

procedure SearchTNC;
begin
  writeln('Searching KISS TNC ...');
end;

procedure TKISS.Decode(frame: string);

function escaped(s: string): string;
var
  i: integer;
  escape: boolean;
  a: char;
begin
  result:='';
  escape:=false;
  for i:=1 to length(s) do
  begin
    a:=s[i];
    if escape then
    begin
      if a=TFEND then result:=result+FEND;
      if a=TFESC then result:=result+FESC;
      escape:=false;
    end
    else
    begin
      if a=FESC then escape:=true
      else result:=result+a;
    end;
  end;
end;

var
  ch,kind: byte;
  len: integer;
begin
  if length(frame)=0 then exit;
  frame:=escaped(frame);
  len:=length(frame);
  kind:=ord(frame[1]) and 15;
  ch:=(ord(frame[1]) shr 4)+1;
  delete(frame,1,1);
  // Do KISS process
  if (kind=0) and (len<256) then URESAT_tlm(@frame[1],len);
end;

procedure TKISS.TimerEvent(Sender: TObject);
begin
  if status=KISS_INACTIVE then Connect;
end;

procedure TKISS.Connect;
begin
  if KISSSocket.ClientSocket.Connected then exit;
  in_buf:='';
  status:=KISS_TRY_CONN;
  try
    KISSSocket.ClientSocket.Connect(host,port);
  except
    status:=KISS_INACTIVE;
  end;
end;

procedure TKISS.Disconnect;
begin
  if status=KISS_ACTIVE then
    if KISSSocket.ClientSocket.Connected then
      try KISSSocket.ClientSocket.Disconnect(true); except end;
end;

constructor TKISS.Create;
begin
  timer:=TTimer.Create(nil);
  timer.Name:='KISSTimer';
  timer.OnTimer:=@TimerEvent;
  timer.Interval:=1000;
  timer.Enabled:=false;
  status:=KISS_INACTIVE;
  in_buf:='';
end;

destructor TKISS.Destroy;
begin
  timer.Enabled:=false;
  timer.Free;
  inherited Destroy;
end;

procedure TKISSSocket.SocketOpen;
begin
  KISS.Connect;
end;

procedure TKISSSocket.SocketClose;
begin
  KISS.Disconnect;
end;

procedure TKISSSocket.OnError(const msg: string; aSocket: TLSocket);
begin
  ErrorCode:=0;
  //Writeln('KISS TNC Link Error');
  KISS.status:=KISS_INACTIVE;
end;

procedure TKISSSocket.OnRead(aSocket: TLSocket);
var
  p: integer;
  msg,frame: string;
begin
  if KISS.status=KISS_ACTIVE then
  begin
    aSocket.GetMessage(msg);
    KISS.in_buf:=KISS.in_buf+msg;
    repeat
      p:=pos(FEND,KISS.in_buf);
      if p>0 then
      begin
        frame:=copy(KISS.in_buf,1,p-1);
        delete(KISS.in_buf,1,p);
        if length(frame)>0 then KISS.Decode(frame);
      end;
    until p=0;
  end;
end;

procedure TKISSSocket.OnConnect(aSocket: TLSocket);
begin
  Writeln('KISS TNC Connected');
  KISS.status:=KISS_ACTIVE;
end;

procedure TKISSSocket.OnDisconnect(aSocket: TLSocket);
begin
  Writeln('KISS TNC Disconnected');
  SearchTNC;
  KISS.status:=KISS_INACTIVE;
end;

constructor TKISSSocket.Create;
begin
  ClientSocket:=TLTCPComponent.Create(nil);
  with ClientSocket do
  begin
    Name:='KISSSocket';
    Host:='';
    Port:=0;
    OnConnect:=@KISSSocket.OnConnect;
    OnDisconnect:=@KISSSocket.OnDisconnect;
    OnReceive:=@KISSSocket.OnRead;
    OnError:=@KISSSocket.OnError;
  end;
end;

destructor TKISSSocket.Destroy;
begin
  KISSSocket.SocketClose;
  ClientSocket.Free;
  inherited Destroy;
end;

procedure ReadIni;
begin
  AppIni:=TIniFile.Create(ExtractFilePath(ParamStr(0))+'kissuresat.ini');
  with AppIni do
  begin
    KISS.host:=ReadString('KISS','Host','localhost');
    KISS.port:=ReadInteger('KISS','Port',8100);
  end;
  AppIni.Free;
end;

procedure WriteIni;
begin
  AppIni:=TIniFile.Create(ExtractFilePath(ParamStr(0))+'kissuresat.ini');
  with AppIni do
  begin
    WriteString('KISS','Host',KISS.host);
    WriteInteger('KISS','Port',KISS.port);
  end;
  AppIni.Free;
end;

procedure AppCreate;
begin
  if KISSSocket=nil then KISSSocket:=TKISSSocket.Create;
  if KISS=nil then KISS:=TKISS.Create;
end;

procedure AppDestroy;
begin
  if KISSSocket<>nil then KISSSocket.Free;
  if KISS<>nil then KISS.Free;
end;

function ConProc(CtrlType: LongWord): LongBool; stdcall;
begin
  case CtrlType of
    CTRL_CLOSE_EVENT, CTRL_C_EVENT, CTRL_BREAK_EVENT, CTRL_LOGOFF_EVENT, CTRL_SHUTDOWN_EVENT:
    begin
      WriteIni;
      AppDestroy;
      Result:=false;
      Exit;
    end;
  end;
  Result:=true;
end;

begin
  SetConsoleTitle(APP_VER);
  AppCreate;
  Application.Initialize;
  Application.CreateForm(TForm,MainForm);
  Application.ShowMainForm:=false;
  SetConsoleCtrlHandler(@ConProc,true);
  ReadIni;
  URESAT_ver;
  SearchTNC;
  KISS.timer.Enabled:=true;
  Application.Run;
  AppDestroy;
end.

